'''
    This section of the codebase is used for construction of route paths using the 
    environmental mesh between a series of user defined waypoints
'''

import copy, json, ast, warnings
import numpy as np
import pandas as pd
from tqdm import tqdm
from shapely import wkt
from shapely.geometry.polygon import Point
import geopandas as gpd
import logging

from pandas.core.common import SettingWithCopyWarning
warnings.simplefilter(action="ignore", category=SettingWithCopyWarning)

from polar_route.crossing import NewtonianDistance
from polar_route.crossing_smoothing import Smoothing,PathValues,find_edge
from polar_route.config_validation.config_validator import validate_route_config
from polar_route.config_validation.config_validator import validate_waypoints

def _flattenCases(id,mesh):
    neighbour_case = []
    neighbour_indx = []
    neighbours = mesh['neighbour_graph'][id]
    for case in neighbours.keys():
        for neighbour in neighbours[case]:
            neighbour_case.append(int(case))
            neighbour_indx.append(int(neighbour))
    return neighbour_case, neighbour_indx

def _initialise_dijkstra_graph(dijkstra_graph):
    '''
        Initialising dijkstra graph information in a standard form

        Args:
            dijkstra_graph (pd.dataframe) - Pandas dataframe of the dijkstra graph construction

        Outputs:
            dijkstra_graph_dict (dict) - Dictionary comprising dijkstra graph with keys based on cellbox id. 
                                         Each entry is a dictionary of the cellbox environmental and dijkstra information. 


    '''

    dijkstra_graph_dict = {}
    for idx,cell in dijkstra_graph.iterrows():
        dijkstra_graph_dict[cell.name] = {}
        dijkstra_graph_dict[cell.name]['id'] = cell.name
        for key in cell.keys():
            entry = cell[key]
            if type(entry) == list:
                entry = np.array(entry)
            dijkstra_graph_dict[cell.name][key] = entry
    return  dijkstra_graph_dict


def _initialise_dijkstra_route(dijkstra_graph,dijkstra_route):
    '''
        Initialising dijkstra route info a standard path form

        Args:
            dijkstra_graph (dict) - Dictionary comprising dijkstra graph with keys based on cellbox id.
                                    Each entry is a dictionary of the cellbox environmental and dijkstra information.

            dijkstra_route (dict) - Dictionary of a GeoJSON entry for the dijkstra route

        Outputs:
            aps (list, [find_edge, ...]) - A list of adjacent cell pairs where each entry is of type find_edge including information on
                                        .crossing, .case, .start, and .end see 'find_edge' for more information
    '''

    org_path_points = np.array(dijkstra_route['geometry']['coordinates'])
    org_cellindices = np.array(dijkstra_route['properties']['CellIndices'])
    org_cellcases= np.array(dijkstra_route['properties']['cases'])

    # -- Generating a dataframe of the case information -- 
    Points      = np.concatenate([org_path_points[0,:][None,:],org_path_points[1:-1:2],org_path_points[-1,:][None,:]])
    cellIndices = np.concatenate([[org_cellindices[0]],[org_cellindices[0]],org_cellindices[1:-1:2],[org_cellindices[-1]],[org_cellindices[-1]]])
    cellcases = np.concatenate([[org_cellcases[0]],[org_cellcases[0]],org_cellcases[1:-1:2],[org_cellcases[-1]],[org_cellcases[-1]]])

    cellDijk    = [dijkstra_graph[ii] for ii in cellIndices]
    cells  = cellDijk[1:-1]
    cases  = cellcases[1:-1]
    aps = []
    for ii in range(len(cells)-1):
        aps += [find_edge(cells[ii],cells[ii+1],cases[ii+1])]

    # #-- Setting some backend information
    start_waypoint = Points[0,:]
    end_waypoint   = Points[-1,:]

    return aps, start_waypoint,end_waypoint

def _json_str(input):
    '''
        Load JSON object either from dict or from file

        Input:
            input (dict or string) - JSON file/dict 
    
        Output:
            output (dict) - Dictionary from JSON object
    '''
    if type(input) is dict:
        output = input
    elif type(input) is str:
        try:
            with open(input, 'r') as f:
                output = json.load(f)
        except:
            raise Exception("Unable to load '{}', please check path name".format(input))
    return output

def _pandas_dataframe_str(input):
    if (type(input) is dict) or (type(input) is pd.core.frame.DataFrame):
        output = input
    elif type(input) is str:
        try:
            output = pd.read_csv(input)
        except:
            raise Exception("Unable to load '{}', please check path name".format(input))
    return output





class RoutePlanner:
    """
        ---

        RoutePlanner optimises the route paths between a series of waypoints. 
        The routes are constructed in a two stage process:

        compute_routes: uses a mesh based Dijkstra method to determine the optimal routes 
                        between a series of waypoint.

        compute_smoothed_routes: smooths the compute_routes using information from the environmental mesh
                                to determine mesh independent optimal route paths


        ---

        Attributes:
            waypoints (dict): A dictionary of the waypoints supplied by the user in the region
                of interest. The waypoints are of the form:

               {\n
                    "Name":{\n
                        '0':"Falklands",\n
                        '1':"Rothera",\n
                        ...\n
                    },\n
                    "Lat":{\n
                        '0':-52.6347222222,
                        '1':-75.26722,\n
                        ...\n
                    },\n
                    "Long":{\n
                        ...\n
                    },\n
                    "index":{\n
                        ...\n
                    }\n
               }



            paths (geojson): A GeoJSON of all paths constructed. The paths are in the form:

                {\n
                    'types':'FeatureCollection',\n
                    "features":{[\n
                        'type':'feature',\n
                        'geometry':{\n
                            'type': 'LineString',

                            'coordinates': [[-27.21694, -75.26722],\n
                                            [-27.5, -75.07960297382266],\n
                                            [-27.619238882768894, -75.0],\n
                                            ...]\n
                        },
                        'properties':{\n
                            'from': 'Halley',\n
                            'to': 'Rothera',\n
                            'traveltime': [0.0,\n
                                        0.03531938671648596,\n
                                        0.050310986633880575,\n
                                        ...],\n
                            'fuel': [0.0,\n
                                    0.9648858923588642,\n
                                    1.3745886107069096,\n
                                    ...],\n
                            'times': ['2017-01-01 00:00:00',
                                    '2017-01-01 00:50:51.595036800',
                                    '2017-01-01 01:12:26.869276800',
                                    ...]\n
                        }\n
                    ]}\n
                }\n

        ---

    """

    def __init__(self, mesh, config, waypoints, cost_func=NewtonianDistance):

        """
            Constructs the routes from information given in the config file.

            Args:

                mesh (dict or string of filepath): mesh based JSON containing the cellbox information and neighbourhood graph

                config (dict or string of filepath): config JSON which defines the attributes required for the route construction. 
                    Sections required for the route construction are as follows\n
                    \n
                    {\n
                        "objective_function": (string) currently either 'traveltime' or 'fuel',\n
                        "path_variables": list of (string),\n
                        "waypoints_path": (string),\n
                        "source_waypoints": list of (string),\n
                        "end_waypoints": list of (string),\n
                        "vector_names": list of (string),\n
                        "zero_currents": (boolean),\n
                        "variable_speed" (boolean),\n
                        "time_unit" (string),\n
                        "early_stopping_criterion" (boolean),\n
                        "save_dijkstra_graphs": (boolean),\n
                        "smooth_path":{\n
                            "max_iteration_number":(int),\n
                            "minimum_difference":(float),\n
                         }\n
                    }\n

                cost_func (func): Crossing point cost function for Dijkstra Path creation. For development purposes only!
        """
        validate_route_config(config)
        validate_waypoints(waypoints)
        # Load in the current cell structure & Optimisation Info̦
        self.mesh             = _json_str(mesh)
        self.config           = _json_str(config)
        waypoints_df          = _pandas_dataframe_str(waypoints)
        source_waypoints_df   = waypoints_df[waypoints_df['Source'] == "X"]
        des_waypoints_df      = waypoints_df[waypoints_df['Destination'] == "X"]

        self.source_waypoints = list(source_waypoints_df['Name'])
        self.end_waypoints    = list(des_waypoints_df['Name'])

        # Case indices
        self.indx_type = np.array([1, 2, 3, 4, -1, -2, -3, -4])

        # Creating a blank path construct
        self.paths          = None
        self.smoothed_paths = None
        self.dijkstra_info = {}

        # ====== Loading Mesh & Neighbour Graph ======
        # Formatting the Mesh and Neighbour Graph to the right form
        self.neighbour_graph = pd.DataFrame(self.mesh['cellboxes']).set_index('id')
        self.neighbour_graph['geometry'] = self.neighbour_graph['geometry'].apply(wkt.loads)
        self.neighbour_graph = gpd.GeoDataFrame(self.neighbour_graph, crs='EPSG:4326', geometry='geometry')

        # Removing any point not in accessible positions
        self.neighbour_graph = self.neighbour_graph.loc[list(self.mesh['neighbour_graph'].keys())]

        # Using neighbour graph to determine neighbour indices and cases
        self.neighbour_graph['case'], self.neighbour_graph['neighbourIndex'] = zip(*self.neighbour_graph.apply(lambda row: _flattenCases(row.name, self.mesh), axis=1))

        self.neighbour_graph.index = self.neighbour_graph.index.astype(int)

        # Renaming the vector columns
        self.neighbour_graph = self.neighbour_graph.rename(columns={self.config['vector_names'][0]: "Vector_x", 
                                                                    self.config['vector_names'][1]: "Vector_y"})

        # ====== Speed Function Checking ======
        # Checking if Speed defined in file
        if 'speed' not in self.neighbour_graph:
            print(self.neighbour_graph.columns)
            raise Exception('Vessel Speed not in the mesh information ! Please run vessel performance')
        
        # ======= Sea-Ice Concentration ======
        if 'SIC' not in self.neighbour_graph:
            self.neighbour_graph['SIC'] = 0.0

        # ====== Objective Function Information ======
        #  Checking if objective function is in the dijkstra            
        if self.config['objective_function'] != 'traveltime':
            if self.config['objective_function'] not in self.neighbour_graph:
                raise Exception("Objective Function require '{}' column in mesh dataframe".format(self.config['objective_function']))

        # ===== Dijkstra Graph =====
        # Adding the required columns needed for the dijkstra graph
        self.neighbour_graph['positionLocked']          = False
        for vrbl in self.config['path_variables']:
            self.neighbour_graph['shortest_{}'.format(vrbl)]    = np.inf
        self.neighbour_graph['neighbourTravelLegs']     = [list() for x in range(len(self.neighbour_graph.index))]
        self.neighbour_graph['neighbourCrossingPoints'] = [list() for x in range(len(self.neighbour_graph.index))]
        self.neighbour_graph['pathIndex']               = [list() for x in range(len(self.neighbour_graph.index))]
        for vrbl in self.config['path_variables']:
            self.neighbour_graph['path_{}'.format(vrbl)] = [list() for x in range(len(self.neighbour_graph.index))]
        self.neighbour_graph['pathPoints']               = [list() for x in range(len(self.neighbour_graph.index))]

        # ====== Defining the cost function ======
        self.cost_func       = cost_func

        # ====== Outlining some constant values ======
        self.unit_time       = self.config['time_unit']
        self.zero_currents   = self.config['zero_currents']
        self.variable_speed  = self.config['variable_speed']
        if type(self.variable_speed) == float:
            logging.info(' Defining a constant speed map = {}'.format(self.variable_speed))
            self.neighbour_graph['speed'] = self.variable_speed
            cbxs = pd.DataFrame(self.mesh['cellboxes'])
            cbxs['speed'] = self.variable_speed
            self.mesh['cellboxes'] = cbxs.to_dict('records')

        # ====== Waypoints ======
        self.mesh['waypoints'] = waypoints_df

        # Initialising Waypoints positions and cell index
        wpts = self.mesh['waypoints']
        wpts['index'] = np.nan
        for idx,wpt in wpts.iterrows():
            indices = self.neighbour_graph[self.neighbour_graph['geometry'].contains(Point(wpt[['Long','Lat']]))].index
            # Waypoint is not within a mesh cell, but could still be on the edge of one. So perturbing the position slightly to the north-east and checking again. 
            #If this is not the case then the waypoint is not within the navigable domain and will continue
            if len(indices) == 0:
                try:
                    indices = mesh[(mesh['geometry'].contains(Point(wpt[['Long','Lat']]+1e-5)))].index
                except:
                    continue
            if len(indices) == 0:
                continue
            if len(indices) > 1:
                raise Exception('Waypoint lies in multiple cell boxes. Please check mesh ! ')
            else:
                wpts['index'].loc[idx] = int(indices[0])

        self.mesh['waypoints'] = wpts[~wpts['index'].isnull()]
        self.mesh['waypoints']['index'] = self.mesh['waypoints']['index'].astype(int)
        self.mesh['waypoints'] =  self.mesh['waypoints'].to_json()

        # ==== Printing Configuration and Information
        self.mesh['waypoints'] =  pd.read_json(self.mesh['waypoints'])

        # # ===== Running the route planner for the given information
        # if ("dijkstra_only" in self.config) and self.config['dijkstra_only']:
        #     self.compute_routes()
        # else:
        #     self.compute_routes()
        #     self.compute_smoothed_routes()


        # # === Saving file to output or saving it to variable output
        # output = self.to_json()
        # if ('output' in self.config) and (type(self.config['output']) == str):
        #     with open(self.config['output'], 'w') as f:
        #         json.dump(output,f)
        # else:
        #     self.output = output



    def to_json(self):
        '''
            Outputting the information in JSON format
        '''
        mesh = copy.copy(self.mesh)
        mesh['waypoints'] = mesh['waypoints'].to_dict()
        output_json = json.loads(json.dumps(mesh))
        del mesh
        return output_json

    def _dijkstra_paths(self, start_waypoints, end_waypoints):
        """
            Hidden function. Given internal variables and start and end waypoints this function
            returns a GEOJSON formatted path dict object

            INPUTS:
                start_waypoints: Start waypoint names (list)
                end_waypoints: End waypoint names (list)
        """

        geojson = dict()
        geojson['type'] = "FeatureCollection"

        paths = []
        wpts_s = self.mesh['waypoints'][self.mesh['waypoints']['Name'].isin(start_waypoints)]
        wpts_e = self.mesh['waypoints'][self.mesh['waypoints']['Name'].isin(end_waypoints)]


        for _, wpt_a in wpts_s.iterrows():
            wpt_a_name  = wpt_a['Name']
            wpt_a_index = int(wpt_a['index'])
            wpt_a_loc   = [[wpt_a['Long'],wpt_a['Lat']]]
            for _, wpt_b in wpts_e.iterrows():
                wpt_b_name  = wpt_b['Name']
                wpt_b_index = int(wpt_b['index'])
                wpt_b_loc   = [[wpt_b['Long'],wpt_b['Lat']]]
                if not wpt_a_name == wpt_b_name:
                    try:
                        graph = self.dijkstra_info[wpt_a_name]
                        path = dict()
                        path['type'] = "Feature"
                        path['geometry'] = {}
                        path['geometry']['type'] = "LineString"
                        path_points = (np.array(wpt_a_loc+list(np.array(graph['pathPoints'].loc[wpt_b_index])[:-1, :])+wpt_b_loc))
                        path['geometry']['coordinates'] = path_points.tolist()

                        path['properties'] = {}
                        path['properties']['name'] = 'Route Path - {} to {}'.format(wpt_a_name, wpt_b_name)
                        path['properties']['from'] = '{}'.format(wpt_a_name)
                        path['properties']['to'] = '{}'.format(wpt_b_name)

                        cellIndices  = np.array(graph['pathIndex'].loc[wpt_b_index])
                        path_indices = np.array([cellIndices[0]] + list(np.repeat(cellIndices[1:-1], 2)) + [cellIndices[-1]])
                        path['properties']['CellIndices'] = path_indices.tolist()

                        cases = []

                        # Determine cases for cell pairs along the path
                        for idx in range(len(cellIndices) -1):
                            cellStart = graph.loc[cellIndices[idx]]
                            cellEnd = graph.loc[cellIndices[idx+1]]
                            case = cellStart['case'][np.where(np.array(cellStart['neighbourIndex']) == cellEnd.name)[0][0]]
                            cases+=[case]

                        start_case = cases[0]
                        end_case = cases[-1]

                        # Full list of cases for each leg (centre to crossing point and crossing point to centre)
                        path_cases = list(np.repeat(cases, 2))

                        # Applying in-cell correction for travel-time
                        cost_func    = self.cost_func(source_graph=self.dijkstra_info[wpt_a_name].loc[path_indices[0]],
                                                      neighbour_graph=self.dijkstra_info[wpt_a_name].loc[path_indices[0]],
                                                      unit_shipspeed='km/hr', unit_time=self.unit_time, zerocurrents=self.zero_currents,
                                                      case=start_case)
                        tt_start = cost_func.waypoint_correction(path_points[0, :], path_points[1, :])
                        cost_func    = self.cost_func(source_graph=self.dijkstra_info[wpt_a_name].loc[path_indices[-1]],
                                                      neighbour_graph=self.dijkstra_info[wpt_a_name].loc[path_indices[0]],
                                                      unit_shipspeed='km/hr', unit_time=self.unit_time, zerocurrents=self.zero_currents,
                                                      case=end_case)

                        tt_end = cost_func.waypoint_correction(path_points[-1, :], path_points[-2, :])
                        path['properties']['traveltime']     = np.array(graph['path_traveltime'].loc[wpt_b_index])
                        path['properties']['traveltime']     = (path['properties']['traveltime'] - path['properties']['traveltime'][0]) + tt_start
                        path['properties']['traveltime'][-1] = (path['properties']['traveltime'][-2] + tt_end)
                        path['properties']['cases'] = [int(p) for p in path_cases]

                        for vrbl in self.config['path_variables']:
                            if vrbl == 'traveltime':
                                continue
                            traveltime_diff = np.r_[path['properties']['traveltime'][0], np.diff(path['properties']['traveltime'])]
                            variable_vals = graph.loc[path_indices, '{}'.format(vrbl)]
                            case_vals = pd.Series(data=[v[np.where(self.indx_type==path_cases[i])[0][0]] for i, v in enumerate(variable_vals)],
                                                  index=variable_vals.index)
                            variable_diff = traveltime_diff*case_vals
                            path['properties'][vrbl] = np.cumsum(variable_diff.to_numpy()).tolist()
                            path['properties']['total_'+vrbl] = path['properties'][vrbl][-1]
                        path['properties']['traveltime'] = path['properties']['traveltime'].tolist()
                        path['properties']['total_traveltime'] = path['properties']['traveltime'][-1]
                        paths.append(path)

                    except:
                        logging.warning('{} to {} - Failed to construct path direct in the dijkstra information'.format(wpt_a_name,wpt_b_name))

        geojson['features'] = paths
        return geojson



    def _objective_value(self, variable, source_graph, neighbour_graph, traveltime, case):
        """
            Hidden variable. Returns the objective value between two cellboxes.
        """
        if variable == 'traveltime':
            objs = np.array([source_graph['shortest_traveltime'] + traveltime[0],source_graph['shortest_traveltime'] + np.sum(traveltime)])
            return objs
        else:
            if type(source_graph['{}'.format(variable)]) == list and len(source_graph['{}'.format(variable)]) != 1:
                idx = np.where(self.indx_type==case)[0][0]
                objs = np.array([source_graph['shortest_{}'.format(variable)] + traveltime[0]*source_graph['{}'.format(variable)][idx],source_graph['shortest_{}'.format(variable)] + traveltime[0]*source_graph['{}'.format(variable)][idx] + traveltime[1]*neighbour_graph['{}'.format(variable)][idx]])
                return objs
            else:
                objs = np.array([source_graph['shortest_{}'.format(variable)] + traveltime[0]*source_graph['{}'.format(variable)], source_graph['shortest_{}'.format(variable)] + traveltime[0]*source_graph['{}'.format(variable)] + traveltime[1]*neighbour_graph['{}'.format(variable)]])
                return objs

    def _neighbour_cost(self, wpt_name, minimum_objective_index):
        """
            Determines the neighbour cost from a source cellbox to all of its neighbours.
            These are then used to update the edge values in the dijkstra graph.
        """
        # Determining the nearest neighbour index for the cell
        source_graph   = self.dijkstra_info[wpt_name].loc[minimum_objective_index]
        # Reset these arrays to empty so no duplicates are produced
        source_graph['neighbourTravelLegs'] = []
        source_graph['neighbourCrossingPoints'] = []

        # Looping over idx
        for idx in range(len(source_graph['case'])):
            indx = source_graph['neighbourIndex'][idx]

            neighbour_graph = self.dijkstra_info[wpt_name].loc[indx]
            case = source_graph['case'][idx]

            # Applying Newton curve to determine crossing point
            cost_func    = self.cost_func(source_graph=source_graph, neighbour_graph=neighbour_graph, case=case,
                                          unit_shipspeed='km/hr', unit_time=self.unit_time,
                                          zerocurrents=self.zero_currents)
            # Updating the Dijkstra graph with the new information
            traveltime, crossing_points,cell_points,case = cost_func.value()

            source_graph['neighbourTravelLegs'].append(traveltime)
            source_graph['neighbourCrossingPoints'].append(np.array(crossing_points))

            # Using neighbourhood cost determine objective function value
            value = self._objective_value(self.config['objective_function'], source_graph,neighbour_graph, traveltime, case)
            if value[1] < neighbour_graph['shortest_{}'.format(self.config['objective_function'])]:
                for vrbl in self.config['path_variables']:
                    value = self._objective_value(vrbl, source_graph, neighbour_graph,traveltime, case)
                    neighbour_graph['shortest_{}'.format(vrbl)] = value[1]
                    neighbour_graph['path_{}'.format(vrbl)]   = source_graph['path_{}'.format(vrbl)] + list(value)
                neighbour_graph['pathIndex']  = source_graph['pathIndex']  + [indx]
                neighbour_graph['pathPoints'] = source_graph['pathPoints'] + [list(crossing_points)] + [list(cell_points)]
                self.dijkstra_info[wpt_name].loc[indx] = neighbour_graph

        self.dijkstra_info[wpt_name].loc[minimum_objective_index] = source_graph

    def _dijkstra(self, wpt_name):
        """
            Runs dijkstra across the whole of the domain.
        """
        # Including only the End Waypoints defined by the user
        wpts = self.mesh['waypoints'][self.mesh['waypoints']['Name'].isin(self.end_waypoints)]
        
        # Initialising zero traveltime at the source location
        source_index = int(self.mesh['waypoints'][self.mesh['waypoints']['Name'] == wpt_name]['index'])

        for vrbl in self.config['path_variables']:
            self.dijkstra_info[wpt_name].loc[source_index, 'shortest_{}'.format(vrbl)] = 0.0
        self.dijkstra_info[wpt_name].loc[source_index, 'pathIndex'].append(source_index)
        
        # # Updating Dijkstra as long as all the waypoints are not visited or for full graph
        if self.config['early_stopping_criterion']:
            stopping_criterion_indices = wpts['index']
        else:
            stopping_criterion_indices = self.dijkstra_info[wpt_name].index

        while (self.dijkstra_info[wpt_name].loc[stopping_criterion_indices, 'positionLocked'] == False).any():

            # Determining the index of the minimum objective function that has not been visited
            minimum_objective_index = self.dijkstra_info[wpt_name][self.dijkstra_info[wpt_name]['positionLocked'] == False]['shortest_{}'.format(self.config['objective_function'])].idxmin()
  
            # Finding the cost of the nearest neighbours from the source cell (Sc)
            self._neighbour_cost(wpt_name,minimum_objective_index)

            # Updating Position to be locked
            self.dijkstra_info[wpt_name].loc[minimum_objective_index, 'positionLocked'] = True

    def compute_routes(self):
        """
            Computes the Dijkstra Paths between waypoints. 
            `waypoints` and `paths` are appended to output JSON
        """

        # Subsetting the waypoints
        if len(self.source_waypoints) == 0:
            raise Exception('No source waypoints defined that are accessible')
        if len(self.end_waypoints) == 0:
            raise Exception('No destination waypoints defined that are accessible')

        # Initialising the Dijkstra Info Dictionary
        for wpt in self.source_waypoints:
            self.dijkstra_info[wpt] = copy.copy(self.neighbour_graph)

        # 
        logging.info('============= Dijkstra Path Creation ============')
        logging.info(' - Objective = {} '.format(self.config['objective_function']))

        for wpt in self.source_waypoints:
            logging.info('--- Processing Waypoint = {} ---'.format(wpt))
            if len(self.mesh['waypoints'][self.mesh['waypoints']['Name'] == wpt]) == 0:
                logging.warning('{} not in accessible waypoints, continuing'.format(wpt))
                continue
            elif (len(self.mesh['waypoints']['Name'] == wpt) > 0) and (len(self.mesh['waypoints']['Name'] != wpt) == 0):
                logging.warning('{} is accessible but has no destination waypoints, continuing'.format(wpt))
                continue
            else:
                self._dijkstra(wpt)


        # Using Dijkstra Graph compute path and meta information to all end_waypoints
        self.paths = self._dijkstra_paths(self.source_waypoints, self.end_waypoints)
        self.mesh['paths'] = self.paths

    def compute_smoothed_routes(self,blocked_metric='SIC',debugging=False):
        """
            Using the previously constructed Dijkstra paths smooth the paths to remove mesh features 
            `paths` will be updated in the output JSON
        """
        # ====== Routes info =======
        if len(self.paths['features']) == 0:
            raise Exception('Paths not constructed as there were no dijkstra paths created')
        routes = copy.deepcopy(self.paths)['features']  


        # ====== Determining route info ======
        if 'smoothing_max_iterations' in self.config:
            max_iterations = self.config['smoothing_max_iterations']
        else:
            max_iterations = 2000
        if 'smoothing_blocked_sic' in self.config:
            blocked_sic = self.config['smoothing_blocked_sic']
        else:
            blocked_sic = 10.0
        if 'smoothing_merge_separation' in self.config:
            merge_separation = self.config['smoothing_merge_separation']
        else:
            merge_separation = 1e-3
        if 'smoothing_converged_sep' in self.config:
            converged_sep = self.config['smoothing_converged_sep']
        else:
            converged_sep = 1e-3

        logging.info('========= Determining Smoothed Paths ===========\n')
        geojson = {}
        SmoothedPaths = []
        for route in routes:
            logging.info('---Smoothing {}'.format(route['properties']['name']))
            dijkstra_graph = self.dijkstra_info[route['properties']['from']]
            self.initialise_dijkstra_graph = _initialise_dijkstra_graph(dijkstra_graph)
            self.route = route
            self.adjacent_pairs,self.start_waypoint,self.end_waypoint = _initialise_dijkstra_route(self.initialise_dijkstra_graph,self.route)

            sf = Smoothing(self.initialise_dijkstra_graph,
                           self.adjacent_pairs,
                           self.start_waypoint,self.end_waypoint,
                           blocked_metric=blocked_metric,
                           max_iterations=max_iterations,
                           blocked_sic = blocked_sic,
                           merge_separation=merge_separation,
                           converged_sep=converged_sep)
            self.sf = sf
            self.sf.forward()

            
            # ------ Smooth Path Values -----
            # Given a smoothed route path now determine the along path parameters.
            pv             = PathValues()
            path_info      = pv.objective_function(sf.aps,sf.start_waypoint,sf.end_waypoint)
            variables      = path_info['variables']
            TravelTimeLegs = variables['traveltime']['path_values']
            DistanceLegs   = variables['distance']['path_values'] 
            pathIndex      = variables['cell_index']['path_values'] 
            FuelLegs       = variables['fuel']['path_values'] 
            SpeedLegs      = variables['speed']['path_values'] 



            # ------ Saving Output in a standard form to be saved ------
            SmoothedPath ={}
            SmoothedPath['type'] = 'Feature'
            SmoothedPath['geometry'] = {}
            SmoothedPath['geometry']['type'] = "LineString"
            SmoothedPath['geometry']['coordinates'] = path_info['path'].tolist()            
            SmoothedPath['properties'] = {}
            SmoothedPath['properties']['from']       = route['properties']['from']
            SmoothedPath['properties']['to']         = route['properties']['to']
            SmoothedPath['properties']['traveltime'] = list(TravelTimeLegs)
            SmoothedPath['properties']['total_traveltime'] = SmoothedPath['properties']['traveltime'][-1]
            SmoothedPath['properties']['fuel']       = list(FuelLegs)
            SmoothedPath['properties']['total_fuel'] = SmoothedPath['properties']['fuel'][-1]
            SmoothedPath['properties']['distance']   = list(DistanceLegs)
            SmoothedPath['properties']['speed']      = list(SpeedLegs)
            SmoothedPaths += [SmoothedPath]

        geojson['type'] = "FeatureCollection"
        geojson['features'] = SmoothedPaths
        self.smoothed_paths = geojson
        self.mesh['paths'] = self.smoothed_paths


import argparse
import json
import inspect
import logging

from meshiphi.mesh_generation.mesh_builder import MeshBuilder

from polar_route import __version__ as version
from polar_route.utils import setup_logging, timed_call, convert_decimal_days
from polar_route.vessel_performance.vessel_performance_modeller import VesselPerformanceModeller
from polar_route.route_planner import RoutePlanner
from polar_route.route_calc import route_calc


@setup_logging
def get_args(
        default_output: str,
        config_arg: bool = True,
        mesh_arg: bool = False,
        waypoints_arg: bool = False
        ):
    """
    Adds required command line arguments to all CLI entry points.

    Args:
        default_output (str): The default output file location.
        config_arg (bool): True if the CLI entry point requires a <config.json> file. Default is True.
        mesh_arg (bool): True if the CLI entry point requires a <mesh.json> file. Default is False.
        waypoints_arg (bool): True if the CLI entry point requires a <waypoints.csv> file. Default is False.

    Returns:

    """
    ap = argparse.ArgumentParser()

    ap.add_argument('--version', action='version',
                        version='%(prog)s {version}'.format(version=version))

    # Optional arguments used in all CLI entry points
    ap.add_argument("-o", "--output",
                    default=default_output,
                    help="Output file")
    ap.add_argument("-v", "--verbose",
                    default=False,
                    action="store_true",
                    help="Turn on DEBUG level logging")

    if config_arg:
        ap.add_argument("config", type=argparse.FileType("r"), 
                    help="File location of a <config.json> file")

    if mesh_arg:
        ap.add_argument("mesh", type=argparse.FileType("r"),
                    help="File location of the environmental mesh")

    if waypoints_arg:
        ap.add_argument("waypoints", type=argparse.FileType("r"))

        # Optional arguments used when route planning.
        ap.add_argument("-p", "--path_geojson",
                        default=False,
                        action = "store_true",
                        help="Output the calculated paths as GeoJSON")

        ap.add_argument("-d", "--dijkstra",
                        default=False,
                        action = "store_true",
                        help="Output dijkstra paths")

    return ap.parse_args()


@timed_call
def resimulate_vehicle_cli():
    """
        CLI entry point for rebuilding the mesh based on its encoded config files.
    """

    default_output = "resimulate_vehicle_output.vessel.json"
    
    args = get_args(default_output, mesh_arg=True, config_arg=False)
    logging.info("{} {}".format(inspect.stack()[0][3][:-4], version))

    mesh_json = json.load(args.mesh)
    mesh_config = mesh_json['config']['mesh_info']

    # Rebuilding mesh, since pruned CB's don't exist in input file
    rebuilt_mesh = MeshBuilder(mesh_config).build_environmental_mesh()
    rebuilt_mesh_json = rebuilt_mesh.to_json()

    # Resimulating vessel
    vessel_config = mesh_json['config']['vessel_info']
    vp = VesselPerformanceModeller(rebuilt_mesh_json, vessel_config)
    vp.model_accessibility()
    vp.model_performance()
    rebuilt_mesh_json = vp.to_json()

    # Saving output
    logging.info("Saving mesh to {}".format(args.output))
    json.dump(rebuilt_mesh_json, open(args.output, "w"), indent=4)


@timed_call
def add_vehicle_cli():
    """
        CLI entry point for the vessel performance modeller
    """

    default_output = "add_vehicle_output.vessel.json"
    args = get_args(default_output, config_arg=True, mesh_arg=True)
    logging.info("{} {}".format(inspect.stack()[0][3][:-4], version))

    mesh_json = json.load(args.mesh)
    vessel_config = json.load(args.config)

    vp = VesselPerformanceModeller(mesh_json, vessel_config)
    vp.model_accessibility()
    vp.model_performance()

    logging.info("Saving mesh to {}".format(args.output))
    info = vp.to_json()
    json.dump(info, open(args.output, "w"), indent=4)


@timed_call
def optimise_routes_cli():
    """
        CLI entry point for the route optimisation
    """
    args = get_args("optimise_routes_output.route.json",
                    config_arg=True, mesh_arg=True ,waypoints_arg= True)
    logging.info("{} {}".format(inspect.stack()[0][3][:-4], version))

    rp = RoutePlanner(args.mesh.name, args.config.name, args.waypoints.name)
    
    output_file = args.output
    output_file_strs = output_file.split('.')
    
    logging.info("Calculating dijkstra routes")
    rp.compute_routes()
    info_dijkstra = rp.to_json()
    
    
    if args.dijkstra:
        # Form a unique name for the dijkstra output
        dijkstra_output_file_strs = output_file_strs
        dijkstra_output_file_strs[0] += '_dijkstra'
        
        logging.info("\tOutputting dijkstra path")
        dijkstra_output_file = '.'.join(dijkstra_output_file_strs)
        json.dump(info_dijkstra, open(dijkstra_output_file, 'w'), indent=4)
        # Create GeoJSON filename
        if args.path_geojson:
            dijkstra_output_file_strs[-1] = 'geojson'
            dijkstra_output_file = '.'.join(dijkstra_output_file_strs)
            logging.info("\tExtracting standalone path GeoJSON")
            json.dump(info_dijkstra['paths'], open(dijkstra_output_file, 'w'), indent=4)
    
    logging.info("Calculating smoothed routes")
    rp.compute_smoothed_routes()
    info = rp.to_json()

    logging.info("Outputting smoothed path")
    json.dump(info, open(output_file, 'w'), indent=4)
    if args.path_geojson:
        # Create GeoJSON filename
        output_file_strs[-1] = 'geojson'
        output_file = '.'.join(output_file_strs)
        logging.info("Extracting standalone path GeoJSON")
        json.dump(info['paths'], open(output_file, 'w'), indent=4)
    
        

@timed_call
def calculate_route_cli():
    """
        CLI entry point to calculate the cost of a manually defined route within an existing mesh.
    """
    args = get_args("calculated_route.json",
                    config_arg = False, mesh_arg = True, waypoints_arg = True)
    logging.info("{} {}".format(inspect.stack()[0][3][:-4], version))

    logging.info(f"Calculating the cost of route {args.waypoints.name} from mesh {args.mesh.name}")

    calc_route = route_calc(args.waypoints.name, args.mesh.name)

    if calc_route is not None:
        max_time = convert_decimal_days(calc_route["features"][0]["properties"]["traveltime"][-1])
        max_fuel = round(calc_route["features"][0]["properties"]["fuel"][-1],2)

        logging.info(f"Calculated route has travel time: {max_time} and fuel cost: {max_fuel} tons")

        logging.info(f"Saving calculated route to {args.output}")
        with open(args.output, "w") as f:
            json.dump(calc_route, f, indent=4)


from polar_route.vessel_performance.abstract_vessel import AbstractVessel
from meshiphi.mesh_generation.aggregated_cellbox import AggregatedCellBox
from abc import abstractmethod
import logging


class AbstractGlider(AbstractVessel):
    """
        Abstract class to model the performance of an underwater glider
    """
    def __init__(self, params):
        """
            Args:
                params (dict): vessel parameters from the vessel config file
        """
        self.vessel_params = params
        logging.info(f"Initialising a vessel object of type: {self.vessel_params['vessel_type']}")
        self.max_speed = self.vessel_params['max_speed']
        self.speed_unit = self.vessel_params['unit']
        self.max_elevation = -1 * self.vessel_params['min_depth']
        self.max_ice = self.vessel_params['max_ice_conc']
        self.excluded_zones = self.vessel_params.get('excluded_zones')


    def model_performance(self, cellbox):
        """
            Method to determine the performance characteristics for the underwater glider

            Args:
                    cellbox (AggregatedCellBox): input cell from environmental mesh
        """
        logging.debug(
            f"Modelling performance in cell {cellbox.id} for a vessel of type: {self.vessel_params['vessel_type']}")
        perf_cellbox = self.model_speed(cellbox)
        perf_cellbox = self.model_battery(perf_cellbox)

        performance_values = {k: v for k, v in perf_cellbox.agg_data.items() if k not in cellbox.agg_data}

        return performance_values

    def model_accessibility(self, cellbox):
        """
            Method to determine if a given cell is accessible to the underwater glider

            Args:
                cellbox (AggregatedCellBox): input cell from environmental mesh
            Returns:
                access_values (dict): boolean values for the modelled accessibility criteria
        """
        logging.debug(f"Modelling accessibility in cell {cellbox.id} for a vessel of type: "
                      f"{self.vessel_params['vessel_type']}")
        access_values = dict()

        # Exclude cells due to land or ice
        access_values['land'] = self.land(cellbox)
        access_values['shallow'] = self.shallow(cellbox)
        access_values['ext_ice'] = self.extreme_ice(cellbox)

        # Exclude any other cell types specified in config
        if self.excluded_zones is not None:
            for zone in self.excluded_zones:
                access_values[zone] = cellbox.agg_data[zone]

        access_values['inaccessible'] = any(access_values.values())

        return access_values

    def land(self, cellbox):
        """
            Method to determine if a cell is land based on sea level

            Args:
                cellbox (AggregatedCellBox): input cell from environmental mesh
            Returns:
                land (bool): boolean that is True if the cell is inaccessible due to land
        """
        if 'elevation' not in cellbox.agg_data:
            logging.warning(f"No elevation data in cell {cellbox.id}, cannot determine if it is land")
            land = False
        else:
            land = cellbox.agg_data['elevation'] >= 0.0

        return land

    def shallow(self, cellbox):
        """
            Method to determine if the water in a cell is too shallow for a glider based on configured minimum depth

            Args:
                cellbox (AggregatedCellBox): input cell from environmental mesh
            Returns:
                shallow (bool): boolean that is True if the cell is too shallow for a glider
        """
        if 'elevation' not in cellbox.agg_data:
            logging.warning(f"No elevation data in cell {cellbox.id}, cannot determine if it is too shallow")
            shallow = False
        else:
            shallow = 0.0 > cellbox.agg_data['elevation'] > self.max_elevation

        return shallow

    def extreme_ice(self, cellbox):
        """
            Method to determine if a cell is inaccessible based on configured max ice concentration

            Args:
                cellbox (AggregatedCellBox): input cell from environmental mesh

            Returns:
                ext_ice (bool): boolean that is True if the cell is inaccessible due to ice
        """
        if 'SIC' not in cellbox.agg_data:
            logging.debug(f"No sea ice concentration data in cell {cellbox.id}")
            ext_ice = False
        else:
            ext_ice = cellbox.agg_data['SIC'] > self.max_ice

        return ext_ice

    @abstractmethod
    def model_speed(self, cellbox: AggregatedCellBox):
        """
            Method to determine the maximum speed that the glider can traverse the given cell

            Args:
                cellbox (AggregatedCellBox): input cell from environmental mesh

            Returns:
                cellbox (AggregatedCellBox): updated cell with speed values
        """
        raise NotImplementedError

    @abstractmethod
    def model_battery(self, cellbox: AggregatedCellBox):
        """
            Method to determine the battery consumption rate of the glider in a given cell

            Args:
                cellbox (AggregatedCellBox): input cell from environmental mesh

            Returns:
                cellbox (AggregatedCellBox): updated cell with battery consumption values
        """
        raise NotImplementedError





import json
import logging
import numpy as np
import pandas as pd
import geopandas as gpd
from shapely import wkt
from shapely.geometry import Point, LineString, MultiLineString, Polygon


# Define ordering of cases in array data
case_indices = np.array([1, 2, 3, 4, -1, -2, -3, -4])

def traveltime_in_cell(xdist, ydist, u, v, s):
    """
        Calculate travel time inside cell.

        Args:
            xdist (float): x distance
            ydist (float): y distance
            u (float): current x component
            v (float): current y component
            s (float): ship speed

        Returns:
            traveltime (float): the travel time
            distance (float): the distance
    """

    dist = np.sqrt(xdist ** 2 + ydist ** 2)
    cval = np.sqrt(u ** 2 + v ** 2)

    dotprod = xdist * u + ydist * v
    diffsqrs = s ** 2 - cval ** 2

    if diffsqrs == 0.0:
        if dotprod == 0.0:
            return np.inf
        else:
            if ((dist ** 2) / (2 * dotprod)) < 0:
                return np.inf
            else:
                traveltime = dist * dist / (2 * dotprod)
                return traveltime

    traveltime = (np.sqrt(dotprod ** 2 + (dist ** 2) * diffsqrs) - dotprod) / diffsqrs
    if traveltime < 0:
        traveltime = np.inf
    return traveltime, dist


def traveltime_distance(cellbox, wp, cp, speed='speed', vector_x='uC', vector_y='vC', case=0):
    """
        Calculate travel time and distance for two points.

        Args:
            cellbox (dict): the cell containing the line segment
            wp (ndarray): the start point
            cp (ndarray): the end point
            speed (str): the key for speed
            vector_x (str): the key for the x vector component
            vector_y (str): the key for the y vector component
            case (int): case giving the index of the speed array

        Returns:
            traveltime (float): the time to travel the line segment
            distance (float): the distance along the line segment
    """

    idx = np.where(case_indices==case)[0][0]
    # Conversion factors from lat/long degrees to metres TODO: replace as part of route planner refactor
    m_long = 111.321 * 1000
    m_lat = 111.386 * 1000
    x = (cp[0] - wp[0]) * m_long * np.cos(wp[1] * (np.pi / 180))
    y = (cp[1] - wp[1]) * m_lat
    su = cellbox[vector_x]
    sv = cellbox[vector_y]
    ssp = cellbox[speed][idx] * (1000 / (60 * 60))
    traveltime, distance = traveltime_in_cell(x, y, su, sv, ssp)
    return traveltime, distance


def case_from_angle(start, end):
    """
        Determine the direction of travel between two points and return the associated case

        Args:
            start (ndarray): the coordinates of the start point within the cell
            end (ndarray):  the coordinates of the end point within the cell

        Returns:
            case (int): the case to use to select variable values from a list
    """

    direct_vec = [end[0]-start[0], end[1]-start[1]]
    direct_ang = np.degrees(np.arctan2(direct_vec[0], direct_vec[1]))

    case = None

    # Angular ranges corresponding to directional cases used in the route planner
    # Cases cover 45 degree angular segments running clockwise from 1 at NE through to -4 at N
    # See Sec. 6.1.2 in the docs for more info

    if -22.5 <= direct_ang < 22.5:
        case = -4
    elif 22.5 <= direct_ang < 67.5:
        case = 1
    elif 67.5 <= direct_ang < 112.5:
        case = 2
    elif 112.5 <= direct_ang < 157.5:
        case = 3
    elif 157.5 <= abs(direct_ang) <= 180:
        case = 4
    elif -67.5 <= direct_ang < -22.5:
        case = -3
    elif -112.5 <= direct_ang < -67.5:
        case = -2
    elif -157.5 <= direct_ang < -112.5:
        case = -1

    return case


def load_route(route_file):
    """
        Load route information from file

        Args:
            route_file (str): Path to user defined route

        Returns:
            df (Dataframe): Dataframe with route info
            from_wp (str): Name of start waypoint
            to_wp (str) Name of end waypoint

    """
    # Loading route from csv file
    if route_file[-3:] == "csv":
        df = pd.read_csv(route_file)
        to_wp = df['Name'].iloc[-1]
        from_wp = df['Name'].iloc[0]
    # Loading route from geojson file
    elif route_file[-4:] == "json":
        with open(route_file, "r") as f:
            route_json = json.load(f)
        route_coords = route_json['features'][0]['geometry']['coordinates']
        to_wp = route_json['features'][0]['properties']['to']
        from_wp = route_json['features'][0]['properties']['from']
        longs = [c[0] for c in route_coords]
        lats = [c[1] for c in route_coords]
        df = pd.DataFrame()
        df['Long'] = longs
        df['Lat'] = lats
    else:
        logging.warning("Invalid route input! Please supply either a csv or geojson file with the route waypoints.")
        return None

    logging.debug(f"Route has {len(df)} waypoints")
    df['id'] = 1
    df['order'] = np.arange(len(df))

    return df, from_wp, to_wp


def load_mesh(mesh_file):
    """
        Load mesh from file into GeoDataFrame
        Args:
            mesh_file (str): Path to mesh with vehicle information

        Returns:
            mesh (GeoDataFrame): Mesh in GeoDataFrame format
    """
    # Loading mesh information
    with open(mesh_file, 'r') as fp:
        info = json.load(fp)
    mesh = pd.DataFrame(info['cellboxes'])
    mesh['geometry'] = mesh['geometry'].apply(wkt.loads)
    mesh = gpd.GeoDataFrame(mesh, crs='EPSG:4326', geometry='geometry')

    region = info['config']['mesh_info']['region']
    region_poly = Polygon(((region['long_min'], region['lat_min']), (region['long_min'], region['lat_max']),
                           (region['long_max'], region['lat_max']), (region['long_max'], region['lat_min'])))

    return mesh, region_poly


def find_intersections(df, mesh):
    """
        Find crossing points of the route and the cells in the mesh
        Args:
            df (DataFrame): Route info in dataframe format
            mesh (GeoDataFrame): Mesh in GeoDataFrame format

        Returns:
            track_points (dict): Dictionary of crossing points and cell ids
    """

    track_waypoints = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df['Long'], df['Lat']))
    tracks = track_waypoints.sort_values(by=['order']).groupby(['id'])['geometry'].apply(
        lambda x: LineString(x.tolist()))
    tracks = gpd.GeoDataFrame(tracks, crs='EPSG:4326', geometry='geometry')

    line_segs_first_points = []
    line_segs_last_points = []
    line_segs_mid_points = []
    line_segs_cell_id = []

    # Find crossing points of the route and the cells in the mesh
    for idx in range(len(mesh)):
        # Skip cells with no intersection
        if not tracks['geometry'].iloc[0].intersects(mesh['geometry'].iloc[idx]):
            continue
        tp = tracks['geometry'].iloc[0].intersection(mesh['geometry'].iloc[idx])
        # Check for multi line strings in case of complex intersections
        if type(tp) == MultiLineString:
            pnts = []
            for l in tp.geoms:
                for c in l.coords:
                    pnts.append(Point(c))
        else:
            pnts = [Point(point) for point in tp.coords]
        if len(pnts) <= 1:
            continue
        line_segs_first_points.append(pnts[0])
        line_segs_mid_points.append(pnts[1:-1])
        line_segs_last_points.append(pnts[-1])
        line_segs_cell_id.append(idx)

    track_points = pd.DataFrame(
        {'startPoints': line_segs_first_points, 'midPoints': line_segs_mid_points, 'endPoints': line_segs_last_points,
         'cellID': line_segs_cell_id})

    return track_points


def order_track(df, track_points):
    """
        Order crossing points into a track along the route
        Args:
            df (DataFrame): Route info in dataframe format
            track_points (dict): Dictionary of crossing points and cell ids

        Returns:
            user_track (DataFrame): DataFrame of ordered crossing points and cell ids
    """

    start_point = Point(df.iloc[0]['Long'], df.iloc[0]['Lat'])
    end_point = Point(df.iloc[-1]['Long'], df.iloc[-1]['Lat'])

    pathing = True
    track_id = np.where(track_points['startPoints'] == start_point)[0][0]
    path_point = []
    cell_ids = []

    # Loop through crossing points to order them into a track along the route
    while pathing:
        try:
            start_point_segment = track_points['startPoints'].iloc[track_id]
            end_point_segment = track_points['endPoints'].iloc[track_id]
            path_point.append(start_point_segment)
            cell_ids.append(track_points['cellID'].iloc[track_id])

            if len(track_points['midPoints'].iloc[track_id]) != 0:
                for midpnt in track_points['midPoints'].iloc[track_id]:
                    path_point.append(midpnt)
                    cell_ids.append(track_points['cellID'].iloc[track_id])

            if end_point_segment == end_point:
                pathing = False
            track_id = np.where(track_points['startPoints'] == end_point_segment)[0][0]
        except IndexError:
            pathing = False
            path_point.append(end_point_segment)
            cell_ids.append('NaN')

    user_track = pd.DataFrame({'Point': path_point, 'CellID': cell_ids})
    return user_track


def route_calc(route_file, mesh_file):
    """
        Function to calculate the fuel/time cost of a user defined route in a given mesh

        Args:
            route_file (str): Path to user defined route
            mesh_file (str): Path to mesh with vehicle information

        Returns:
            user_path (dict): User defined route in geojson format with calculated cost information
    """

    # Load route info and waypoint names from file
    df, from_wp, to_wp = load_route(route_file)

    # Load mesh info from file
    mesh, region_poly = load_mesh(mesh_file)

    # Check route waypoints contained in mesh bounds
    for idx in range(len(df)):
        if region_poly.contains(Point((df.iloc[idx]['Long'],df.iloc[idx]['Lat']))):
            continue
        else:
            logging.warning(f"Mesh does not contain waypoint located at Lat: {df.iloc[idx]['Lat']} "
                            f"Long: {df.iloc[idx]['Long']} !")
            return None

    # Find points where route crosses mesh
    track_points = find_intersections(df, mesh)

    # Loop through crossing points to order them into a track along the route
    user_track = order_track(df, track_points)
    logging.debug(f"Route has {len(user_track)} crossing points")

    # Initialise segment costs with zero values at start point of path
    traveltimes = [0.0]
    distances = [0.0]
    cellboxes = [mesh.iloc[user_track['CellID'].iloc[0]]]
    cases = [1]

    # Calculate cost of each segment in the path
    for idx in range(len(user_track)-1):
        start_point = np.array((user_track['Point'].iloc[idx].xy[0][0], user_track['Point'].iloc[idx].xy[1][0]))
        end_point = np.array((user_track['Point'].iloc[idx+1].xy[0][0], user_track['Point'].iloc[idx+1].xy[1][0]))
        cell_box = mesh.iloc[user_track['CellID'].iloc[idx]]
        case = case_from_angle(start_point, end_point)
        # Check for inaccessible cells on user defined route
        if cell_box['inaccessible']:
            logging.warning(f"This route crosses an inaccessible cell! Cell located at Lat: {cell_box['cy']} "
                         f"Long: {cell_box['cx']}. Please reroute around it.")
            return None

        traveltime_s, distance_m = traveltime_distance(cell_box, start_point, end_point, speed='speed', vector_x='uC',
                                                   vector_y='vC', case=case)
        traveltime = ((traveltime_s / 60) / 60) / 24
        distance = distance_m / 1000
        traveltimes.append(traveltime)
        distances.append(distance)
        cellboxes.append(cell_box)
        cases.append(case)


    logging.debug(f"Route crosses {len(set([c['id'] for c in cellboxes]))} different cellboxes")
    # Find cumulative values along path
    path_points = user_track['Point']
    path_traveltimes = np.cumsum(traveltimes)
    path_distances = np.cumsum(distances)
    path_fuels = [traveltimes[idx] * cellboxes[idx]['fuel'][np.where(case_indices==cases[idx])[0][0]] for idx in range(len(traveltimes))]
    path_fuel = np.cumsum(path_fuels)

    # Put path values into geojson format
    path_geojson = pd.DataFrame()
    path_geojson['geometry'] = [LineString(path_points)]
    path_geojson['traveltime'] = [path_traveltimes.tolist()]
    path_geojson['distance'] = [(path_distances * 1000).tolist()]
    path_geojson['fuel'] = [path_fuel.tolist()]
    path_geojson['to'] = to_wp
    path_geojson['from'] = from_wp
    path_geojson = gpd.GeoDataFrame(path_geojson, crs='EPSG:4326', geometry='geometry')

    user_path = json.loads(path_geojson.to_json())

    return user_path


from meshiphi.mesh_generation.aggregated_cellbox import AggregatedCellBox
from abc import ABCMeta, abstractmethod

class AbstractVessel(metaclass=ABCMeta):
    """
    Interface to define the abstract methods required for any vessel class to work within the VesselPerformanceModeller.
    """

    @abstractmethod
    def __init__(self, params: dict):
        """
        Initialise the vessel object with parameters from the config.

        Args:
            params (dict): vessel parameters from the vessel config file
        """
        raise NotImplementedError

    @abstractmethod
    def model_performance(self, cellbox: AggregatedCellBox):
        """
        Calculate performance parameters for the given vessel.

        Args:
            cellbox (AggregatedCellBox): cell in which performance is being modelled

        Returns:
            performance_values (dict): values for relevant performance parameters
        """
        raise NotImplementedError

    @abstractmethod
    def model_accessibility(self, cellbox: AggregatedCellBox):
        """
        Determine accessibility of the input cell for the given vessel.

        Args:
            cellbox (AggregatedCellBox): cell in which accessibility is being determined

        Returns:
            access_values (dict): values for the accessibility and other related booleans
        """
        raise NotImplementedError



__version__ = "0.3.9"
__description__ = "PolarRoute: Long-distance maritime polar route planning taking into account complex changing environmental conditions"
__license__ = "MIT"
__author__ = "Jonathan Smith, Samuel Hall, George Coombs, Harrison Abbot, Ayat Fekry, James Byrne, Michael Thorne, Maria Fox"
__email__ = "polarroute@bas.ac.uk"
__copyright__ = "2022-2023, BAS AI Lab"

# Wrapped in try-except so that setup.py can import polar_route without crashing due to dependency errors
try:
    from meshiphi.mesh_generation.mesh_builder import MeshBuilder as MeshBuilder
    from meshiphi.dataloaders.factory import DataLoaderFactory as DataLoaderFactory
    from meshiphi.mesh_generation.boundary import Boundary as Boundary

    from polar_route.vessel_performance.vessel_performance_modeller import VesselPerformanceModeller as VesselPerformanceModeller

    from polar_route.route_planner import RoutePlanner as RoutePlanner

except ModuleNotFoundError as err:
    print(f'{err}\n Is PolarRoute installed correctly?')



from polar_route.vessel_performance.vessels.SDA import SDA
from polar_route.vessel_performance.vessels.slocum import SlocumGlider

class VesselFactory:
    """
        Factory class to produce initialised vessel objects.
    """
    @classmethod
    def get_vessel(cls, config):
        """
            Method to return an initialised instance of a vessel class designed for performance modelling

            Args:
                config (dict): a vessel config dictionary

            Returns:
                vessel: an instance of a vessel class designed for performance modelling
        """
        vessel_requirements = {"SDA": (SDA, ["max_speed", "unit", "beam", "hull_type", "force_limit", "max_ice_conc",
                                             "min_depth"]),
                               "Slocum": (SlocumGlider, ["max_speed", "unit", "max_ice_conc", "min_depth"])}

        vessel_type = config['vessel_type']

        if vessel_type in vessel_requirements:
            vessel_class = vessel_requirements[vessel_type][0]
            required_params = vessel_requirements[vessel_type][1]
        else:
            raise ValueError(f'{vessel_type} not in known list of vessels')

        assert all(key in config for key in required_params), \
            f'Dataloader {vessel_type} is missing some parameters! Requires {required_params}. Has {list(config.keys())}'

        vessel = vessel_class(config)

        return vessel



from meshiphi.mesh_generation.aggregated_cellbox import AggregatedCellBox
from polar_route.vessel_performance.vessels.abstract_ship import AbstractShip
import numpy as np
import logging

class SDA(AbstractShip):
    """
        Vessel class with methods specifically designed to model the performance of the British Antarctic Survey
        research and supply ship, the RRS Sir David Attenborough (SDA)
    """
    def __init__(self, params):
        """
            Args:
                params (dict): vessel parameters from the vessel config file
        """
        super().__init__(params)

        self.force_limit = self.vessel_params['force_limit']
        self.beam = self.vessel_params['beam']
        self.hull_type = self.vessel_params['hull_type']

    def model_speed(self, cellbox):
        """
            Method to determine the maximum speed that the SDA can traverse the given cell

            Args:
                cellbox (AggregatedCellBox): input cell from environmental mesh

            Returns:
                cellbox (AggregatedCellBox): updated cell with speed values
        """

        logging.debug(f"Calculating new speed for cellbox {cellbox.id} based on SDA resistance models")
        speed = cellbox.agg_data['speed']
        ice_resistance = None

        if all(k in cellbox.agg_data for k in ("SIC", "thickness", "density")):
            logging.debug("Adjusting speed according to ice resistance model")
            if cellbox.agg_data['SIC'] == 0.0:
                ice_resistance = 0.
                speed = self.max_speed
            elif cellbox.agg_data['SIC'] > self.max_ice:
                speed = 0.0
                ice_resistance = np.inf
            else:
                ice_resistance = self.ice_resistance(cellbox)
                if ice_resistance > self.force_limit:
                    speed = self.invert_resistance(cellbox)
                    cellbox.agg_data['speed'] = speed
                    ice_resistance = self.ice_resistance(cellbox)
                else:
                    speed = self.max_speed
        else:
            logging.debug("No resistance data available, no speed adjustment necessary")

        logging.debug("Creating speed array")
        cellbox.agg_data['speed'] = [speed for x in range(8)]

        if ice_resistance is not None:
            cellbox.agg_data['ice resistance'] = ice_resistance

        return cellbox

    def model_fuel(self, cellbox):
        """
            Method to determine the fuel consumption rate of the SDA in a given cell

            Args:
                cellbox (AggregatedCellBox): input cell from environmental mesh

            Returns:
                cellbox (AggregatedCellBox): updated cell with fuel consumption values
        """
        logging.debug(f"Calculating fuel requirements in cell {cellbox.id}")

        cellbox = self.model_resistance(cellbox)

        cellbox.agg_data['fuel'] = [fuel_eq(cellbox.agg_data['speed'][i], r)
                                    for i, r in enumerate(cellbox.agg_data['resistance'])]
        return cellbox


    def model_resistance(self, cellbox):
        """
            Method to determine the resistance force acting on the SDA in a given cell

            Args:
                cellbox (AggregatedCellBox): input cell from environmental mesh

            Returns:
                cellbox (AggregatedCellBox): updated cell with resistance values
        """

        # Include ice resistance if present
        if 'ice resistance' in cellbox.agg_data:
            ice_resistance = [cellbox.agg_data['ice resistance'] for x in range(8)]
        else:
            ice_resistance = [0, 0, 0, 0, 0, 0, 0, 0]

        # Calculate wind resistance if wind data present:
        if 'u10' in cellbox.agg_data and 'v10' in cellbox.agg_data:
            cellbox = calc_wind(cellbox)
            cellbox.agg_data['resistance'] = [cellbox.agg_data['wind resistance'][i] + ice_resistance[i] for i in range(8)]
        else:
            logging.debug("No wind data present, wind resistance will not be calculated")
            cellbox.agg_data['resistance'] = ice_resistance

        return cellbox

    def ice_resistance(self, cellbox):
        """
            Method to find the ice resistance force acting on the SDA at a given speed in a given cell

            The input cellbox should contain the following values:
                velocity (float): The speed of the vessel in km/h
                sic (float): The average sea ice concentration in the cell as a percentage
                thickness (float): The average ice thickness in the cell in m
                density (float): The average ice density in the cell in kg/m^3

            Args:
                cellbox (AggregatedCellBox): input cell from environmental mesh

            Returns:
                resistance (float): Resistance force in N
        """

        velocity = cellbox.agg_data['speed']
        sic = cellbox.agg_data['SIC']
        thickness = cellbox.agg_data['thickness']
        density = cellbox.agg_data['density']

        # If there's no ice then return zero
        if not sic:
            return 0.

        # Model parameters for different hull types
        hull_params = {'slender': [4.4, -0.8267, 2.0], 'blunt': [16.1, -1.7937, 3]}

        hull = self.hull_type
        beam = self.beam
        kparam, bparam, nparam = hull_params[hull]
        gravity = 9.81  # m/s-2

        speed = velocity * (5. / 18.)  # assume km/h and convert to m/s

        froude = speed / np.sqrt(gravity * (sic / 100) * thickness)
        resistance = 0.5 * kparam * (froude ** bparam) * density * beam * thickness * (speed ** 2) * (
                    (sic / 100) ** nparam)
        return resistance

    def invert_resistance(self, cellbox):
        """
            Method to find the vessel speed that keeps the ice resistance force below a given threshold in a given cell

            The input cellbox should contain the following values\n
                sic (float) - The average sea ice concentration in the cell as a percentage \n
                thickness (float) - The average ice thickness in the cell in m \n
                density (float) - The average ice density in the cell in kg/m^3 \n

            Args:
                cellbox (AggregatedCellBox): input cell from environmental mesh

            Returns:
                new_speed (float): Safe vessel speed in km/h
        """

        sic = cellbox.agg_data['SIC']
        thickness = cellbox.agg_data['thickness']
        density = cellbox.agg_data['density']

        # If there's no ice then return max speed
        if not sic:
            return self.max_speed

        # Model parameters for different hull types
        hull_params = {'slender': [4.4, -0.8267, 2.0], 'blunt': [16.1, -1.7937, 3]}
        # force_limit (float): Resistance force value that should not be exceeded in N
        force_limit = self.force_limit

        hull = self.hull_type
        beam = self.beam
        kparam, bparam, nparam = hull_params[hull]
        gravity = 9.81  # m/s-2

        vexp = 2 * force_limit / (kparam * density * beam * thickness * ((sic / 100) ** nparam) * (
                gravity * thickness * sic / 100) ** -(bparam / 2))

        vms = vexp ** (1 / (2.0 + bparam))
        new_speed = vms * (18. / 5.)  # convert from m/s to km/h

        return new_speed

    def wave_resistance(self, w_height):
        """
        Method to calculate the wave resistance given the wave height and vessel geometry.
        Recommended by the ITTC for small wave heights: https://ittc.info/media/1936/75-04-01-012.pdf
        """
        rho_w = 9807 # N/m^3 (specific weight of water at 4°C from wikipedia, will vary with temp and salinity)
        beam = self.vessel_params['Beam']
        c_block = self.vessel_params.get('c_block', 0.75) # ratio of underwater volume to cuboid
        length = self.vessel_params['Length']

        wave_res = (0.64*rho_w*c_block*(w_height**2)*(beam**2))/length # Kreitner, valid up to ~2m wave height

        return wave_res


def fuel_eq(speed, resistance):
    """
        Equation to calculate the fuel consumption in tons/day given the speed in km/h and the resistance force in N

        Args:
            speed (float): the SDA's speed in km/h
            resistance (float): the resistance force in N

        Returns:
            fuel (float): the fuel consumption in tons/day
    """
    # Assume no effect from "negative resistance" (e.g. when tailwind is stronger than ice resistance)
    if resistance < 0:
        resistance = 0.
    fuel = (0.00137247 * speed ** 2 - 0.0029601 * speed + 0.25290433
          + 7.75218178e-11 * resistance ** 2 + 6.48113363e-06 * resistance) * 24.0
    return fuel


def c_wind(rel_ang):
    """
        Function to return the wind resistance coefficient for some relative angle between wind and travel directions.
    """
    cs = -1 * np.array([-0.94, -0.77, -0.42, -0.48, -0.17, 0.30, 0.34])
    angles = np.array([0., np.pi / 6., np.pi / 3., np.pi / 2., 2 * np.pi / 3., 5 * np.pi / 6., np.pi])
    return np.interp(rel_ang, angles, cs)


def wind_mag_dir(cellbox, va):
    """
        Function that returns the relative wind speed and direction given the speed and heading of the vessel
        and the easterly and northerly components of the true wind vector. Speeds in m/s.
    """
    vs = cellbox.agg_data['speed'][0]*0.277778
    uw = cellbox.agg_data['u10']
    vw = cellbox.agg_data['v10']

    # Define wind vector
    w_vec = np.array([uw, vw])

    # Find vessel vector in component form
    uv, vv = vs * np.sin(va), vs * np.cos(va)
    v_vec = np.array([uv, vv])

    # Find apparent wind vector
    aw_vec = w_vec - v_vec

    # Find magnitude of apparent wind
    ws = np.linalg.norm(aw_vec)

    # Define unit vectors for dot product
    if vs:
        unit_v = v_vec / vs
    else:
        unit_v = [0., 0.]
    if ws:
        unit_aw = aw_vec / ws
    else:
        return 0., 0.

    # Calculate dot product and find angle
    dp = np.dot(unit_v, unit_aw)
    ang = np.arccos(-1 * dp)

    return ws, ang


def wind_resistance(v_speed, w_speed, rel_ang):
    """
        Function to calculate the wind resistance given the wind speed and direction and the vessel speed.
    """
    a = 750.
    rho = 1.225

    wind_res = 0.5 * rho * (w_speed**2) * a * c_wind(rel_ang) - 0.5 * rho * (v_speed ** 2) * a * c_wind(0)

    return wind_res


def calc_wind(cellbox):
    """
        Function to calculate the wind resistance as well as the relative wind speed and angle for a vessel traversing a
        cell with 8 different equally spaced angular headings.

        Args:
            cellbox (AggregatedCellBox): input cell from environmental mesh

        Returns:
            cellbox (AggregatedCellBox): updated cell with wind information
    """

    logging.debug(f"Calculating wind resistance in cellbox {cellbox.id}")

    wind_res = [0, 0, 0, 0, 0, 0, 0, 0]
    rel_wind_speed = [0, 0, 0, 0, 0, 0, 0, 0]
    rel_wind_angle = [0, 0, 0, 0, 0, 0, 0, 0]

    heads = [np.pi/4, np.pi/2, 3*np.pi/4, np.pi, 5*np.pi/4, 3*np.pi/2, 7*np.pi/4, 0.]
    for i, head in enumerate(heads):
        rel_wind_speed[i], rel_wind_angle[i] = wind_mag_dir(cellbox, head)
        wind_res[i] = wind_resistance(cellbox.agg_data['speed'][i]*0.277778, rel_wind_speed[i],
                                           rel_wind_angle[i]) # assume km/h and convert to m/s

    cellbox.agg_data['wind resistance'] = wind_res
    cellbox.agg_data['relative wind speed'] = rel_wind_speed
    cellbox.agg_data['relative wind angle'] = rel_wind_angle

    return cellbox







"""
    The python package `crossing` implement the optimisation for the crossing point for the unsmoothed and smoothed path
    construction. The package is separated into two classes `NewtonianDistance` and `NewtonianCurve`.
    In the section below we will go through, stage by stage, how the crossing point is determined and the methods
    used within the classes.

"""
import copy
import pandas as pd
import numpy as np
import pyproj
import logging
np.seterr(divide='ignore', invalid='ignore')


class NewtonianDistance:

    def __init__(self,source_graph=None,neighbour_graph=None,
                 case=None,unit_shipspeed='km/hr',unit_time='days',
                 zerocurrents=True,debugging=False,maxiter=1000,optimizer_tol=1e-3):
        """Example of docstring on the __init__ method.

        The __init__ method may be documented in either the class level
        docstring, or as a docstring on the __init__ method itself.

        Either form is acceptable, but the two should not be mixed. Choose one
        convention to document the __init__ method and be consistent with it.

        Note:
            Do not include the `self` parameter in the ``Args`` section.

        Args:
            param1 (str): Description of `param1`.
            param2 (:obj:`int`, optional): Description of `param2`. Multiple
                lines are supported.
            param3 (:obj:`list` of :obj:`str`): Description of `param3`.

        """
        # Cell information
        self.source_graph     = source_graph
        self.neighbour_graph  = neighbour_graph

        # Case indices
        self.indx_type = np.array([1, 2, 3, 4, -1, -2, -3, -4])

        # Inside the code the base units are m/s. Changing the units of the inputs to match
        indx = np.where(self.indx_type==case)[0][0]
        self.unit_shipspeed  = unit_shipspeed
        self.unit_time       = unit_time
        self.source_speed    = self._unit_speed(self.source_graph['speed'][indx])
        self.neighbour_speed = self._unit_speed(self.neighbour_graph['speed'][indx])
        self.case            = case

        if zerocurrents:
            self.zero_current_factor = 0.0
        else:
            self.zero_current_factor = 1.0

        # Optimisation Information
        self.maxiter       = maxiter
        self.optimizer_tol = optimizer_tol

        # Optimisation Information
        self.m_long  = 111.321*1000
        self.m_lat   = 111.386*1000

        # Defining a small distance
        self.small_distance = 1e-4

        # For Debugging purposes
        self.debugging     = debugging

    def _newton_optimisation(self, f, x, a, Y, u1, v1, u2, v2, s1, s2):
        '''
            Newton Optimisation applied to the crossing point, to update its crossing position in-line 
            with environmental parameters of the starting and end cell box.

            All inputs and outputs are given in SI units (m,m/s,s)

            Inputs:
                f  (func) - 
                x  (float) - Start cell perpendicular distance from cell centre to crossing boundary
                a  (float) - End cell perpendicular distance from cell centre to crossing boundary
                Y  (float) - Parallel distance, to crossing boundary, between start cell and end cell centres
                u1 (float) - Start Cell perpendicular to crossing boundary forcing component 
                v1 (float) - Start Cell parallel to crossing boundary forcing component 
                u2 (float) - End Cell perpendicular to crossing boundary forcing component 
                v2 (float) - End Cell parallel to crossing boundary forcing component  
                s1 (float) - Start Cell max speed 
                s1 (float) - End Cell max speed

            Returns:
                y0 (float) - updated crossing point represent as parallel distance along crossing boundary from start cell centre to crossing point
                t  (tuple, (float,float)) - Traveltime of the two segments from start cell centre to crossing to end cell centre.
       '''
        y0 = (Y*x)/(x+a)
        improving = True
        iteration_num = 0
        while improving:
            F,dF,X1,X2,t1,t2  = f(y0,x,a,Y,u1,v1,u2,v2,s1,s2)
            if F==np.inf:
                return np.nan,np.inf  
            y0  = y0 - (F/dF)
            improving = abs((F/dF)/(X1*X2)) > self.optimizer_tol
            iteration_num+=1
            # Assume convergence impossible after 1000 iterations and exit
            if iteration_num>1000:
                # Set crossing point to nan and travel times to infinity
                y0 = np.nan
                t1 = -1
                t2 = -1
        return y0,self._unit_time(np.array([t1,t2]))

    def _unit_speed(self,val):
        '''
            Applying unit speed for an input type.
            
            Input:
                Val (float) - Input speed in m/s
            Output:
                Val (float) - Output speed in unit type unit_shipspeed

        '''
        if not isinstance(val,type(None)):
            if self.unit_shipspeed == 'km/hr':
                val = val*(1000/(60*60))
            if self.unit_shipspeed == 'knots':
                val = (val*0.51)
            return val
        else:
            return None

    def _unit_time(self,val):
        '''
            Applying unit time for an input type.
            
            Input:
                Val (float) - Input time in s
            Output:
                Val (float) - Output time in unit type unit_time

        '''
        if self.unit_time == 'days':
            val = val/(60*60*24.)
        elif self.unit_time == 'hr':
            val = val/(60*60.)
        elif self.unit_time == 'min':
            val = val/(60.)
        elif self.unit_time == 's':
            val = val
        return val


    def _F(self,y,x,a,Y,u1,v1,u2,v2,s1,s2):
        '''
            Determining Newton Function and differential of newton function from adjacent cell pair information

            Args:
                y  (float) - Crossing point as a parallel distance along crossing boundary from start cell centre to crossing point
                x  (float) - Start cell perpendicular distance from cell centre to crossing boundary
                a  (float) - End cell perpendicular distance from cell centre to crossing boundary
                Y  (float) - Parallel distance, to crossing boundary, between start cell and end cell centres
                u1 (float) - Start Cell perpendicular to crossing boundary forcing component 
                v1 (float) - Start Cell parallel to crossing boundary forcing component 
                u2 (float) - End Cell perpendicular to crossing boundary forcing component 
                v2 (float) - End Cell parallel to crossing boundary forcing component  
                s1 (float) - Start Cell max speed 
                s1 (float) - End Cell max speed

            Outputs:
                F (float)  - Newton function value
                dF (float) - Differential of Newton function value
                X1 (float) - Characteristic X distance in start cell
                X2 (float) - Characteristic X distance in end cell
                t1 (float) - Traveltime leg from start cell centre to crossing point
                t1 (float) - Traveltime leg from crossing point to end cell centre

        '''
        C1 = s1**2 - u1**2 - v1**2
        C2 = s2**2 - u2**2 - v2**2
        D1 = x*u1 + y*v1
        D2 = a*u2 + (Y-y)*v2
        X1ns = D1**2 + C1*(x**2 + y**2)
        if X1ns < 0:
            X1 = -1
        else:
            X1 = np.sqrt(X1ns)
        X2ns = D2**2 + C2*(a**2 + (Y-y)**2)
        if X2ns < 0:
            X2 = -1
        else:
            X2 = np.sqrt(X2ns)

        if X1 < 0 or X2 < 0:
            return np.inf,np.inf,np.inf,np.inf,np.inf,np.inf

        F  = X2*(y-((v1*(X1-D1))/C1)) + X1*(y-Y+((v2*(X2-D2))/C2))

        dD1 = v1
        dD2 = -v2
        if X1 == 0:
            dX1 = 0
        else:
            dX1 = (D1*v1 + C1*y)/X1
        if X2 ==0:
            dX2 = 0
        else:
            dX2 = (-D2*v2 - C2*(Y-y))/X2
        dF  = (X1+X2) + y*(dX1 + dX2) - (v1/C1)*(dX2*(X1-D1) + X2*(dX1-dD1))\
            + (v2/C2)*(dX1*(X2-D2)+X1*(dX2-dD2)) - Y*dX1

        t1 = (X1-D1)/C1
        t2 = (X2-D2)/C2

        return F,dF,X1,X2,t1,t2

    def _traveltime_in_cell(self,xdist,ydist,U,V,S):
        '''
            Determines the travel-time within cell between two points

            Args:
                xdist (float) - Longitudinal distance in cell in m
                xdist (float) - Latitudinal distance in cell in m
                U (float) - Longitudinal component of forcing vector (m/s)
                V (float) - Latitudinal component of forcing vector (m/s)
                S (float) - Max speed in cell (m/s)


        '''
        dist  = np.sqrt(xdist**2 + ydist**2)
        cval  = np.sqrt(U**2 + V**2)

        dotprod  = xdist*U + ydist*V
        diffsqrs = S**2 - cval**2

        # if (dotprod**2 + diffsqrs*(dist**2) < 0)
        if diffsqrs == 0.0:
            if dotprod == 0.0:
                return np.inf
            else:
                if ((dist**2)/(2*dotprod))  <0:
                    return np.inf
                else:
                    traveltime = dist * dist / (2 * dotprod)
                    return traveltime

        traveltime = (np.sqrt(dotprod**2 + (dist**2)*diffsqrs) - dotprod)/diffsqrs
        if traveltime < 0:
            traveltime = np.inf
        return traveltime

    def waypoint_correction(self,Wp,Cp):
        '''
            Determines the traveltime between two points within a given cell
            Args:
                Wp (tuple): Start Waypoint location (long,lat)
                Cp (tuple): End Waypoint location (long,lat)
            Returns:
                traveltime (float) - Traveltime between the two points within cell in unit_time
        '''
        x = (Cp[0]-Wp[0])*self.m_long*np.cos(Wp[1]*(np.pi/180))
        y = (Cp[1]-Wp[1])*self.m_lat
        Su  = self.source_graph['Vector_x']*self.zero_current_factor
        Sv  = self.source_graph['Vector_y']*self.zero_current_factor
        Ssp = self.source_speed
        traveltime = self._traveltime_in_cell(x,y,Su,Sv,Ssp)
        return self._unit_time(traveltime)


    def _longitude(self):
        '''
            Applying an inplace longitude correction to the crossing point between the start and end cell boxes

            All outputs are given in SI units (m,m/s,s)

            Outputs:
                Traveltime (tuple,[float,float])  - Traveltime legs from start cell centre to crossing to end cell centre
                CrossPoints (tuple,[float,float]) - Crossing Point in (long,lat)
                CellPoints (tuple,[int,int])      - Start and End cell indices

        '''

        if self.case==2:
            ptvl = 1.0
        else:
            ptvl = -1.0

        s_cx  = self.source_graph['cx']
        s_cy  = self.source_graph['cy']
        s_dcx = self.source_graph['dcx']
        s_dcy = self.source_graph['dcy']
        n_cx  = self.neighbour_graph['cx']
        n_cy  = self.neighbour_graph['cy']
        n_dcx = self.neighbour_graph['dcx']
        n_dcy = self.neighbour_graph['dcy']


        Su = ptvl*self.source_graph['Vector_x']*self.zero_current_factor
        Sv = ptvl*self.source_graph['Vector_y']*self.zero_current_factor
        Nu = ptvl*self.neighbour_graph['Vector_x']*self.zero_current_factor
        Nv = ptvl*self.neighbour_graph['Vector_y']*self.zero_current_factor

        Ssp = self.source_speed
        Nsp = self.neighbour_speed


        x = s_dcx*self.m_long*np.cos(s_cy*(np.pi/180))
        a = n_dcx*self.m_long*np.cos(n_cy*(np.pi/180))
        Y = ptvl*(n_cy-s_cy)*self.m_lat

        # Optimising to determine the y-value of the crossing point
        y,TravelTime = self._newton_optimisation(self._F,x,a,Y,Su,Sv,Nu,Nv,Ssp,Nsp)
        if np.isnan(y) or TravelTime[0] < 0 or TravelTime[1] < 0:
            TravelTime  = [np.inf,np.inf]
            CrossPoints = [np.nan,np.nan]
            CellPoints  = [np.nan,np.nan]
            return TravelTime,CrossPoints,CellPoints

        
        CrossPoints = (s_cx+ptvl*s_dcx,\
                       s_cy+ptvl*y/self.m_lat)
        CellPoints  = [n_cx,n_cy]


        # Checking Crossing Point possible
        # Defining the min and max of the start and end cells
        smin = s_cy-s_dcy 
        smax = s_cy+s_dcy 
        emin = n_cy-n_dcy
        emax = n_cy+n_dcy
        vmin = np.max([smin,emin])
        vmax = np.min([smax,emax])
        if (CrossPoints [1] < vmin) or (CrossPoints[1] > vmax):
            CrossPoints = (CrossPoints[0],np.clip(CrossPoints[1],vmin,vmax))


        return TravelTime,CrossPoints,CellPoints

    def _latitude(self):
        '''
            Applying an inplace latitude correction to the crossing point between the start and end cell boxes

            All outputs are given in SI units (m,m/s,s)

            Outputs:
                Traveltime (tuple,[float,float])  - Traveltime legs from start cell centre to crossing to end cell centre
                CrossPoints (tuple,[float,float]) - Crossing Point in (long,lat)
                CellPoints (tuple,[int,int])      - Start and End cell indices

        '''

        if self.case==4:
            ptvl = 1.0
        else:
            ptvl = -1.0


        s_cx  = self.source_graph['cx']
        s_cy  = self.source_graph['cy']
        s_dcx = self.source_graph['dcx']
        s_dcy = self.source_graph['dcy']
        n_cx  = self.neighbour_graph['cx']
        n_cy  = self.neighbour_graph['cy']
        n_dcx = self.neighbour_graph['dcx']
        n_dcy = self.neighbour_graph['dcy']


        Su = -1*ptvl*self.source_graph['Vector_y']*self.zero_current_factor
        Sv = ptvl*self.source_graph['Vector_x']*self.zero_current_factor
        Nu = -1*ptvl*self.neighbour_graph['Vector_y']*self.zero_current_factor
        Nv = ptvl*self.neighbour_graph['Vector_x']*self.zero_current_factor


        Ssp=self.source_speed
        Nsp=self.neighbour_speed

        x = s_dcy*self.m_lat
        a = n_dcy*self.m_lat
        Y = ptvl*(n_cx-s_cx)*self.m_long*np.cos((n_cy+s_cy)*(np.pi/180)/2.0)
        
        
        y,TravelTime   = self._newton_optimisation(self._F,x,a,Y,Su,Sv,Nu,Nv,Ssp,Nsp)
        if np.isnan(y) or TravelTime[0] < 0 or TravelTime[1] < 0:
            TravelTime  = [np.inf,np.inf]
            CrossPoints = [np.nan,np.nan]
            CellPoints  = [np.nan,np.nan]
            return TravelTime,CrossPoints,CellPoints


        clon = s_cx  + ptvl*y/(self.m_long*np.cos((n_cy+s_cy)*(np.pi/180)/2.0))
        clat = s_cy + -1*ptvl*s_dcy

        CrossPoints = (clon,clat)
        CellPoints  = [n_cx,n_cy]


        # Checking Crossing Point possible
        # Defining the min and max of the start and end cells
        smin = s_cx-s_dcx 
        smax = s_cx+s_dcx 
        emin = n_cx-n_dcx
        emax = n_cx+n_dcx
        vmin = np.max([smin,emin])
        vmax = np.min([smax,emax])
        if (CrossPoints [0] < vmin) or (CrossPoints[0] > vmax):
            CrossPoints = (np.clip(CrossPoints[0],vmin,vmax),CrossPoints[1])

        return TravelTime,CrossPoints,CellPoints


    def _corner(self):
        '''
            Applying an inplace corner crossing point between the start and end cell boxes

            All outputs are given in SI units (m,m/s,s)

            Outputs:
                Traveltime (tuple,[float,float])  - Traveltime legs from start cell centre to crossing to end cell centre
                CrossPoints (tuple,[float,float]) - Crossing Point in (long,lat)
                CellPoints (tuple,[int,int])      - Start and End cell indices

        '''


        s_cx  = self.source_graph['cx']
        s_cy  = self.source_graph['cy']
        s_dcx = self.source_graph['dcx']
        s_dcy = self.source_graph['dcy']
        n_cx  = self.neighbour_graph['cx']
        n_cy  = self.neighbour_graph['cy']
        n_dcx = self.neighbour_graph['dcx']
        n_dcy = self.neighbour_graph['dcy']



        # Given the case determine the positive and negative position relative to centre
        if self.case==1:
            ptvX = 1.0
            ptvY = 1.0
        elif self.case==-1:
            ptvX = -1.0
            ptvY = -1.0
        elif self.case==3:
            ptvX = 1.0
            ptvY = -1.0
        elif self.case==-3:
            ptvX = -1.0
            ptvY = 1.0

        dx1 = ptvX*s_dcx*self.m_long*np.cos(s_cy*(np.pi/180))
        dx2 = ptvX*n_dcx*self.m_long*np.cos(n_cy*(np.pi/180))
        dy1 = ptvY*s_dcy*self.m_lat
        dy2 = ptvY*n_dcy*self.m_lat

        # Currents in Cells
        Su = self.source_graph['Vector_x']*self.zero_current_factor
        Sv = self.source_graph['Vector_y']*self.zero_current_factor
        Nu = self.neighbour_graph['Vector_x']*self.zero_current_factor
        Nv = self.neighbour_graph['Vector_y']*self.zero_current_factor



        # Vehicles Speeds in Cells
        Ssp = self.source_speed; Nsp = self.neighbour_speed

        # Determining the crossing point as the corner of the case
        CrossPoints = [s_cx+ptvX*s_dcx,\
                       s_cy+ptvY*s_dcy]
        CellPoints  = [n_cx,n_cy]

        # Determining traveltime
        t1 = self._traveltime_in_cell(dx1,dy1,Su,Sv,Ssp)
        t2 = self._traveltime_in_cell(dx2,dy2,Nu,Nv,Nsp)
        TravelTime  = self._unit_time(np.array([t1,t2]))

        if TravelTime[0] < 0 or TravelTime[1] < 0:
            TravelTime  = [np.inf,np.inf]
            CrossPoints = [np.nan,np.nan]
            CellPoints  = [np.nan,np.nan]
            return TravelTime,CrossPoints,CellPoints


        return TravelTime,CrossPoints,CellPoints


    def value(self):
        '''
            Applying a correction to determine the optimal crossing point between the start and end cell boxes

            All outputs are given in SI units (m,m/s,s)

            Outputs:
                Traveltime (tuple,[float,float])  - Traveltime legs from start cell centre to crossing to end cell centre
                CrossPoints (tuple,[float,float]) - Crossing Point in (long,lat)
                CellPoints (tuple,[int,int])      - Start and End cell indices
                Case (int)                        - Adjacent cell-box case between start and end cell box

        '''
        if abs(self.case)==2:
            TravelTime,CrossPoints,CellPoints = self._longitude()
        elif abs(self.case)==4:
            TravelTime,CrossPoints,CellPoints = self._latitude()
        elif abs(self.case)==1 or abs(self.case)==3:
            TravelTime,CrossPoints,CellPoints = self._corner()
        else:
            TravelTime  = [np.inf,np.inf]
            CrossPoints = [np.nan,np.nan]
            CellPoints  = [np.nan,np.nan]

        return TravelTime, CrossPoints, CellPoints, self.case


import numpy as np
import logging

from meshiphi.mesh_generation.environment_mesh import EnvironmentMesh
from polar_route.vessel_performance.vessel_factory import VesselFactory
from polar_route.config_validation.config_validator import validate_vessel_config
from polar_route.utils import timed_call

class VesselPerformanceModeller:
    """
        Class for modelling the vessel performance.
        Takes both an environmental mesh and vessel config as input in json format and modifies the input mesh to
        include vessel specifics.
    """
    def __init__(self, env_mesh_json, vessel_config):
        """

        Args:
            env_mesh_json (dict): a dictionary loaded from an environmental mesh json file
            vessel_config (dict): a dictionary loaded from a vessel config json file
        """
        logging.info("Initialising Vessel Performance Modeller")
        validate_vessel_config(vessel_config)

        self.env_mesh = EnvironmentMesh.load_from_json(env_mesh_json)
        self.config = vessel_config
        self.vessel = VesselFactory.get_vessel(vessel_config)

        self.filter_nans()

    @timed_call
    def model_accessibility(self):
        """

        Method to determine the accessibility of cells in the environmental mesh and remove inaccessible cells from the
        neighbour graph.

        """
        for i, cellbox in enumerate(self.env_mesh.agg_cellboxes):
            access_values = self.vessel.model_accessibility(cellbox)
            self.env_mesh.update_cellbox(i, access_values)
        inaccessible_nodes = [c.id for c in self.env_mesh.agg_cellboxes if c.agg_data['inaccessible']]
        logging.info(f"Found {len(inaccessible_nodes)} inaccessible cells in the mesh")
        for in_node in inaccessible_nodes:
            self.env_mesh.neighbour_graph.remove_node_and_update_neighbours(in_node)

    @timed_call
    def model_performance(self):
        """

        Method to calculate the relevant vessel performance values for each cell in the environmental mesh and update
        the mesh accordingly.

        """
        for i, cellbox in enumerate(self.env_mesh.agg_cellboxes):
            if cellbox.agg_data.get('inaccessible'):
                continue
            performance_values = self.vessel.model_performance(cellbox)
            self.env_mesh.update_cellbox(i, performance_values)

    def to_json(self):
        """
            Method to return the modified mesh in json format.

            Returns:
                j_mesh (dict): a dictionary representation of the modified mesh.
        """
        self.env_mesh.config['vessel_info'] = self.config
        j_mesh = self.env_mesh.to_json()
        return j_mesh

    def filter_nans(self):
        """
            Method to check for NaNs in the input cell boxes and zero them if present
        """
        for i, cellbox in enumerate(self.env_mesh.agg_cellboxes):
            if any(np.isnan(val) for val in cellbox.agg_data.values() if type(val) == float):
                filtered_data = {k: 0. if np.isnan(v) else v for k, v in cellbox.agg_data.items() if type(v) == float}
                self.env_mesh.update_cellbox(i, filtered_data)


from polar_route.vessel_performance.vessels.abstract_glider import AbstractGlider
from meshiphi.mesh_generation.aggregated_cellbox import AggregatedCellBox
import numpy as np
import logging

class SlocumGlider(AbstractGlider):
    """
        Vessel class with methods specifically designed to model the performance of the Slocum G2 Glider
    """
    def __init__(self, params):
        """
            Args:
                params (dict): vessel parameters from the vessel config file
        """

        super().__init__(params)
        # Coefficients from fit to figures provided by Alex
        speed_coefficients = np.array([4.44444444, -0.5555555499999991])
        self.speed_polynomial = np.poly1d(speed_coefficients)
        depth_coefficients = np.array([0.001, 2])
        self.depth_polynomial = np.poly1d(depth_coefficients)

    def model_speed(self, cellbox):
        """
            Method to determine the maximum speed that the glider can traverse the given cell

            Args:
                cellbox (AggregatedCellBox): input cell from environmental mesh

            Returns:
                cellbox (AggregatedCellBox): updated cell with speed values
        """

        cellbox.agg_data['speed'] = [self.max_speed for x in range(8)]
        return cellbox

    def model_battery(self, cellbox):
        """
            Method to determine the rate of battery usage in a given cell in Ah/day

            Args:
                cellbox (AggregatedCellBox): input cell from environmental mesh

            Returns:
                cellbox (AggregatedCellBox): updated cell with battery consumption values
        """

        if cellbox.agg_data['elevation'] > self.max_elevation:
            battery = np.inf
        elif cellbox.agg_data['elevation'] <= -1000.0:
            # Assume yo-yo dive to 1000m
            # Battery consumption changes only with speed
            battery = [self.speed_polynomial(s) for s in cellbox.agg_data['speed']]
        else:
            # Battery consumption change with speed
            battery_speed = [self.speed_polynomial(s) for s in cellbox.agg_data['speed']]

            # Estimate depth variation based on linear fit to figures from Alex's presentation
            battery = [bs * self.depth_polynomial(cellbox.agg_data['elevation']) for bs in battery_speed]

        cellbox.agg_data['battery'] = battery
        return cellbox
